@@ -1,26 +1,31 @@
 import asyncio
-import binascii
 import logging
 import os
 
-from serial import SerialException
-from zigpy.exceptions import DeliveryError
-from zigpy.quirks import CustomDevice, CustomEndpoint
-from zigpy.types import BroadcastAddress
-import zigpy.application
-import zigpy.device
-import zigpy.util
-import zigpy.zdo
-
-import bellows.types as t
-import bellows.zigbee.util
 from bellows.exception import ControllerError, EzspError
 import bellows.multicast
+import bellows.types as t
+import bellows.zigbee.util
+from serial import SerialException
+import voluptuous as vol
+import zigpy.application
+import zigpy.device
+from zigpy.quirks import CustomDevice, CustomEndpoint
+from zigpy.types import BroadcastAddress
+import zigpy.util
+import zigpy.zdo
+import zigpy.zdo.types as zdo_t
 
 APS_ACK_TIMEOUT = 120
-APS_REPLY_TIMEOUT = 5
-APS_REPLY_TIMEOUT_EXTENDED = 28
+CONF_PARAM_SRC_RTG = "source_routing"
+CONFIG_SCHEMA = zigpy.application.CONFIG_SCHEMA.extend(
+    {vol.Optional(CONF_PARAM_SRC_RTG, default=False): bellows.zigbee.util.cv_boolean}
+)
+EZSP_DEFAULT_RADIUS = 0
+EZSP_MULTICAST_NON_MEMBER_RADIUS = 3
 MAX_WATCHDOG_FAILURES = 4
+MTOR_MIN_INTERVAL = 600
+MTOR_MAX_INTERVAL = 1800
 RESET_ATTEMPT_BACKOFF_TIME = 5
 WATCHDOG_WAKE_PERIOD = 10
 
@@ -30,15 +35,20 @@ LOGGER = logging.getLogger(__name__)
 class ControllerApplication(zigpy.application.ControllerApplication):
     direct = t.EmberOutgoingMessageType.OUTGOING_DIRECT
 
-    def __init__(self, ezsp, database_file=None):
-        super().__init__(database_file=database_file)
+    def __init__(self, ezsp, database_file=None, config={}):
+        super().__init__(database_file=database_file, config=CONFIG_SCHEMA(config))
         self._ctrl_event = asyncio.Event()
         self._ezsp = ezsp
         self._multicast = bellows.multicast.Multicast(ezsp)
-        self._pending = Requests()
+        self._pending = zigpy.util.Requests()
         self._watchdog_task = None
         self._reset_task = None
         self._in_flight_msg = None
+        self._tx_options = t.EmberApsOption(
+            t.EmberApsOption.APS_OPTION_RETRY
+            | t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
+        )
+        self.use_source_routing = self.config[CONF_PARAM_SRC_RTG]
 
     @property
     def controller_event(self):
@@ -67,10 +77,11 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._cfg(c.CONFIG_SECURITY_LEVEL, 5)
         await self._cfg(c.CONFIG_SUPPORTED_NETWORKS, 1)
         zdo = (
-            t.EmberZdoConfigurationFlags.APP_RECEIVES_SUPPORTED_ZDO_REQUESTS |
-            t.EmberZdoConfigurationFlags.APP_HANDLES_UNSUPPORTED_ZDO_REQUESTS
+            t.EmberZdoConfigurationFlags.APP_RECEIVES_SUPPORTED_ZDO_REQUESTS
+            | t.EmberZdoConfigurationFlags.APP_HANDLES_UNSUPPORTED_ZDO_REQUESTS
         )
         await self._cfg(c.CONFIG_APPLICATION_ZDO_FLAGS, zdo)
+        await self._cfg(c.CONFIG_PAN_ID_CONFLICT_REPORT_THRESHOLD, 2)
         await self._cfg(c.CONFIG_TRUST_CENTER_ADDRESS_CACHE_SIZE, 2)
         await self._cfg(c.CONFIG_ADDRESS_TABLE_SIZE, 16)
         await self._cfg(c.CONFIG_SOURCE_ROUTE_TABLE_SIZE, 8)
@@ -78,14 +89,17 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._cfg(c.CONFIG_INDIRECT_TRANSMISSION_TIMEOUT, 7680)
         await self._cfg(c.CONFIG_KEY_TABLE_SIZE, 1)
         await self._cfg(c.CONFIG_TRANSIENT_KEY_TIMEOUT_S, 180, True)
-        await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT, 60)
-        await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT, 8)
-        await self._cfg(c.CONFIG_MULTICAST_TABLE_SIZE,
-                        self.multicast.TABLE_SIZE)
-        await self._cfg(c.CONFIG_PACKET_BUFFER_COUNT, 0xff)
+        if self._ezsp.ezsp_version >= 7:
+            await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT, 8)
+        else:
+            await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT, 60)
+            await self._cfg(c.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT, 8)
+        await self._cfg(c.CONFIG_MULTICAST_TABLE_SIZE, self.multicast.TABLE_SIZE)
+        await self._cfg(c.CONFIG_PACKET_BUFFER_COUNT, 0xFF)
 
         status, count = await e.getConfigurationValue(
-            c.CONFIG_APS_UNICAST_MESSAGE_COUNT)
+            c.CONFIG_APS_UNICAST_MESSAGE_COUNT
+        )
         assert status == t.EmberStatus.SUCCESS
         self._in_flight_msg = asyncio.Semaphore(count)
         LOGGER.debug("APS_UNICAST_MESSAGE_COUNT is set to %s", count)
@@ -94,17 +108,25 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             output_clusters=[zigpy.zcl.clusters.security.IasZone.cluster_id]
         )
 
-    async def add_endpoint(self, endpoint=1,
-                           profile_id=zigpy.profiles.zha.PROFILE_ID,
-                           device_id=0xbeef,
-                           app_flags=0x00,
-                           input_clusters=[],
-                           output_clusters=[]):
+    async def add_endpoint(
+        self,
+        endpoint=1,
+        profile_id=zigpy.profiles.zha.PROFILE_ID,
+        device_id=0xBEEF,
+        app_flags=0x00,
+        input_clusters=[],
+        output_clusters=[],
+    ):
         """Add endpoint."""
         res = await self._ezsp.addEndpoint(
-            endpoint, profile_id, device_id, app_flags,
-            len(input_clusters), len(output_clusters),
-            input_clusters, output_clusters
+            endpoint,
+            profile_id,
+            device_id,
+            app_flags,
+            len(input_clusters),
+            len(output_clusters),
+            input_clusters,
+            output_clusters,
         )
         LOGGER.debug("Ezsp adding endpoint: %s", res)
 
@@ -113,6 +135,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self.initialize()
         e = self._ezsp
 
+        await self.set_source_routing()
         v = await e.networkInit()
         if v[0] != t.EmberStatus.SUCCESS:
             if not auto_form:
@@ -142,8 +165,25 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         self.handle_join(self.nwk, self.ieee, 0)
         LOGGER.debug("EZSP nwk=0x%04x, IEEE=%s", self._nwk, str(self._ieee))
+
         await self.multicast.startup(self.get_device(self.ieee))
 
+    async def set_source_routing(self) -> None:
+        res = await self._ezsp.setConcentrator(
+            self.use_source_routing,
+            t.EmberConcentratorType.HIGH_RAM_CONCENTRATOR,
+            MTOR_MIN_INTERVAL,
+            MTOR_MAX_INTERVAL,
+            2,
+            5,
+            0,
+        )
+        LOGGER.debug("Set concentrator type: %s", res)
+        if res[0] != t.EmberStatus.SUCCESS:
+            LOGGER.warning(
+                "Couldn't set concentrator type %s: %s", self.use_source_routing, res
+            )
+
     async def shutdown(self):
         """Shutdown and cleanup ControllerApplication."""
         LOGGER.info("Shutting down ControllerApplication")
@@ -159,7 +199,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         channel = t.uint8_t(channel)
 
         if pan_id is None:
-            pan_id = t.uint16_t.from_bytes(os.urandom(2), 'little')
+            pan_id = t.uint16_t.from_bytes(os.urandom(2), "little")
         pan_id = t.uint16_t(pan_id)
 
         if extended_pan_id is None:
@@ -191,8 +231,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         """Set up the policies for what the NCP should do"""
         e = self._ezsp
         v = await e.setPolicy(
-            t.EzspPolicyId.TC_KEY_REQUEST_POLICY,
-            t.EzspDecisionId.DENY_TC_KEY_REQUESTS,
+            t.EzspPolicyId.TC_KEY_REQUEST_POLICY, t.EzspDecisionId.DENY_TC_KEY_REQUESTS
         )
         assert v[0] == t.EmberStatus.SUCCESS  # TODO: Better check
         v = await e.setPolicy(
@@ -213,22 +252,44 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
     def ezsp_callback_handler(self, frame_name, args):
         LOGGER.debug("Received %s frame with %s", frame_name, args)
-        if frame_name == 'incomingMessageHandler':
+        if frame_name == "incomingMessageHandler":
             self._handle_frame(*args)
-        elif frame_name == 'messageSentHandler':
+        elif frame_name == "messageSentHandler":
             if args[4] != t.EmberStatus.SUCCESS:
                 self._handle_frame_failure(*args)
             else:
                 self._handle_frame_sent(*args)
-        elif frame_name == 'trustCenterJoinHandler':
+        elif frame_name == "trustCenterJoinHandler":
             if args[2] == t.EmberDeviceUpdate.DEVICE_LEFT:
                 self.handle_leave(args[0], args[1])
             else:
                 self.handle_join(args[0], args[1], args[4])
-        elif frame_name == '_reset_controller_application':
+        elif frame_name == "incomingRouteRecordHandler":
+            self.handle_route_record(*args)
+        elif frame_name == "incomingRouteErrorHandler":
+            self.handle_route_error(*args)
+        elif frame_name == "_reset_controller_application":
             self._handle_reset_request(*args)
 
-    def _handle_frame(self, message_type, aps_frame, lqi, rssi, sender, binding_index, address_index, message):
+    def _handle_frame(
+        self,
+        message_type,
+        aps_frame,
+        lqi,
+        rssi,
+        sender,
+        binding_index,
+        address_index,
+        message,
+    ):
+        if (
+            aps_frame.clusterId == zdo_t.ZDOCmd.Device_annce
+            and aps_frame.destinationEndpoint == 0
+        ):
+            nwk, rest = t.uint16_t.deserialize(message[1:])
+            ieee, _ = t.EmberEUI64.deserialize(rest)
+            LOGGER.info("ZDO Device announce: 0x%04x, %s", nwk, ieee)
+            self.handle_join(nwk, ieee, 0)
         try:
             device = self.get_device(nwk=sender)
         except KeyError:
@@ -236,66 +297,54 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             return
 
         device.radio_details(lqi, rssi)
-        try:
-            tsn, command_id, is_reply, args = self.deserialize(device, aps_frame.sourceEndpoint, aps_frame.clusterId, message)
-        except ValueError as e:
-            LOGGER.error("Failed to parse message (%s) on cluster %d, because %s", binascii.hexlify(message), aps_frame.clusterId, e)
-            return
+        self.handle_message(
+            device,
+            aps_frame.profileId,
+            aps_frame.clusterId,
+            aps_frame.sourceEndpoint,
+            aps_frame.destinationEndpoint,
+            message,
+        )
 
-        if is_reply:
-            self._handle_reply(device, aps_frame, tsn, command_id, args)
-        else:
-            self.handle_message(device, False, aps_frame.profileId, aps_frame.clusterId, aps_frame.sourceEndpoint, aps_frame.destinationEndpoint, tsn, command_id, args)
-
-    @staticmethod
-    def _dst_pp(addr, aps):
-        """Return format string and args."""
-        ep, cluster = aps.destinationEndpoint, aps.clusterId
-        return '[0x%04x:%s:0x%04x]: ', (addr, ep, cluster)
-
-    def _handle_reply(self, sender, aps_frame, tsn, command_id, args):
-        try:
-            request = self._pending[tsn]
-            if request.reply:
-                request.reply.set_result(args)
-            return
-        except KeyError:
-            hdr, hdr_args = self._dst_pp(sender.nwk, aps_frame)
-            LOGGER.debug(hdr + "Unexpected response TSN=%s command=%s args=%s",
-                         *(hdr_args + (tsn, command_id, args)))
-        except asyncio.futures.InvalidStateError as exc:
-            hdr, hdr_args = self._dst_pp(sender.nwk, aps_frame)
-            LOGGER.debug(hdr + "Invalid state on future - probably duplicate response: %s",
-                         *(hdr_args + (exc, )))
-            # We've already handled, don't drop through to device handler
-            return
-
-        self.handle_message(sender, True, aps_frame.profileId, aps_frame.clusterId, aps_frame.sourceEndpoint, aps_frame.destinationEndpoint, tsn, command_id, args)
-
-    def _handle_frame_failure(self, message_type, destination, aps_frame, message_tag, status, message):
-        hdr, hdr_args = self._dst_pp(destination, aps_frame)
+    def _handle_frame_failure(
+        self, message_type, destination, aps_frame, message_tag, status, message
+    ):
         try:
             request = self._pending[message_tag]
-            msg = hdr + "message send failure: %s"
-            msg_args = (hdr_args + (status, ))
-            request.send.set_exception(DeliveryError(msg % msg_args))
+            request.result.set_result((status, "message send failure"))
         except KeyError:
-            LOGGER.debug(hdr + "Unexpected message send failure", *hdr_args)
-        except asyncio.futures.InvalidStateError as exc:
-            LOGGER.debug(hdr + "Invalid state on future - probably duplicate response: %s",
-                         *(hdr_args + (exc, )))
+            LOGGER.debug(
+                "Unexpected message send failure for message tag %s", message_tag
+            )
+        except asyncio.InvalidStateError as exc:
+            LOGGER.debug(
+                (
+                    "Invalid state on future for message tag %s "
+                    "- probably duplicate response: %s"
+                ),
+                message_tag,
+                exc,
+            )
 
-    def _handle_frame_sent(self, message_type, destination, aps_frame, message_tag, status, message):
+    def _handle_frame_sent(
+        self, message_type, destination, aps_frame, message_tag, status, message
+    ):
         try:
             request = self._pending[message_tag]
-            request.send.set_result(True)
+            request.result.set_result(
+                (t.EmberStatus.SUCCESS, "message sent successfully")
+            )
         except KeyError:
-            hdr, hdr_args = self._dst_pp(destination, aps_frame)
-            LOGGER.debug(hdr + "Unexpected message send notification", *hdr_args)
-        except asyncio.futures.InvalidStateError as exc:
-            hdr, hdr_args = self._dst_pp(destination, aps_frame)
-            LOGGER.debug(hdr + "Invalid state on future - probably duplicate response: %s",
-                         *(hdr_args + (exc, )))
+            LOGGER.debug("Unexpected message send notification tag: %s", message_tag)
+        except asyncio.InvalidStateError as exc:
+            LOGGER.debug(
+                (
+                    "Invalid state on future for message tag %s "
+                    "- probably duplicate response: %s"
+                ),
+                message_tag,
+                exc,
+            )
 
     def _handle_reset_request(self, error):
         """Reinitialize application controller."""
@@ -315,8 +364,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 await self._reset_controller()
                 break
             except (asyncio.TimeoutError, SerialException) as exc:
-                LOGGER.debug(
-                    "ControllerApplication reset unsuccessful: %s", str(exc))
+                LOGGER.debug("ControllerApplication reset unsuccessful: %s", str(exc))
             await asyncio.sleep(RESET_ATTEMPT_BACKOFF_TIME)
 
         self._reset_task = None
@@ -329,9 +377,86 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._ezsp.reconnect()
         await self.startup()
 
-    @zigpy.util.retryable_request
-    async def request(self, nwk, profile, cluster, src_ep, dst_ep, sequence, data, expect_reply=True,
-                      timeout=APS_REPLY_TIMEOUT):
+    async def mrequest(
+        self,
+        group_id,
+        profile,
+        cluster,
+        src_ep,
+        sequence,
+        data,
+        *,
+        hops=EZSP_DEFAULT_RADIUS,
+        non_member_radius=EZSP_MULTICAST_NON_MEMBER_RADIUS
+    ):
+        """Submit and send data out as a multicast transmission.
+
+        :param group_id: destination multicast address
+        :param profile: Zigbee Profile ID to use for outgoing message
+        :param cluster: cluster id where the message is being sent
+        :param src_ep: source endpoint id
+        :param sequence: transaction sequence number of the message
+        :param data: Zigbee message payload
+        :param hops: the message will be delivered to all nodes within this number of
+                     hops of the sender. A value of zero is converted to MAX_HOPS
+        :param non_member_radius: the number of hops that the message will be forwarded
+                                  by devices that are not members of the group. A value
+                                  of 7 or greater is treated as infinite
+        :returns: return a tuple of a status and an error_message. Original requestor
+                  has more context to provide a more meaningful error message
+        """
+        if not self.is_controller_running:
+            raise ControllerError("ApplicationController is not running")
+
+        aps_frame = t.EmberApsFrame()
+        aps_frame.profileId = t.uint16_t(profile)
+        aps_frame.clusterId = t.uint16_t(cluster)
+        aps_frame.sourceEndpoint = t.uint8_t(src_ep)
+        aps_frame.destinationEndpoint = t.uint8_t(src_ep)
+        aps_frame.options = t.EmberApsOption(
+            t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
+        )
+        aps_frame.groupId = t.uint16_t(group_id)
+        aps_frame.sequence = t.uint8_t(sequence)
+        message_tag = self.get_sequence()
+
+        with self._pending.new(message_tag) as req:
+            async with self._in_flight_msg:
+                res = await self._ezsp.sendMulticast(
+                    aps_frame, hops, non_member_radius, message_tag, data
+                )
+                if res[0] != t.EmberStatus.SUCCESS:
+                    return res[0], "EZSP sendMulticast failure: %s" % (res[0],)
+
+                res = await asyncio.wait_for(req.result, APS_ACK_TIMEOUT)
+        return res
+
+    async def request(
+        self,
+        device,
+        profile,
+        cluster,
+        src_ep,
+        dst_ep,
+        sequence,
+        data,
+        expect_reply=True,
+        use_ieee=False,
+    ):
+        """Submit and send data out as an unicast transmission.
+
+        :param device: destination device
+        :param profile: Zigbee Profile ID to use for outgoing message
+        :param cluster: cluster id where the message is being sent
+        :param src_ep: source endpoint id
+        :param dst_ep: destination endpoint id
+        :param sequence: transaction sequence number of the message
+        :param data: Zigbee message payload
+        :param expect_reply: True if this is essentially a request
+        :param use_ieee: use EUI64 for destination addressing
+        :returns: return a tuple of a status and an error_message. Original requestor
+                  has more context to provide a more meaningful error message
+        """
         if not self.is_controller_running:
             raise ControllerError("ApplicationController is not running")
 
@@ -340,32 +465,46 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         aps_frame.clusterId = t.uint16_t(cluster)
         aps_frame.sourceEndpoint = t.uint8_t(src_ep)
         aps_frame.destinationEndpoint = t.uint8_t(dst_ep)
-        aps_frame.options = t.EmberApsOption(
-            t.EmberApsOption.APS_OPTION_RETRY |
-            t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
-        )
+        aps_frame.options = self._tx_options
         aps_frame.groupId = t.uint16_t(0)
         aps_frame.sequence = t.uint8_t(sequence)
+        message_tag = self.get_sequence()
 
-        dev = self.get_device(nwk=nwk)
-        if dev.node_desc.is_end_device in (True, None):
-            LOGGER.debug("Extending timeout for %s/0x%04x", dev.ieee, nwk)
-            await self._ezsp.setExtendedTimeout(dev.ieee, True)
-            timeout = APS_REPLY_TIMEOUT_EXTENDED
-        with self._pending.new(sequence, expect_reply) as req:
+        if use_ieee:
+            LOGGER.warning(
+                ("EUI64 addressing is not currently supported, " "reverting to NWK")
+            )
+        with self._pending.new(message_tag) as req:
             async with self._in_flight_msg:
-                res = await self._ezsp.sendUnicast(self.direct, nwk, aps_frame,
-                                                   sequence, data)
+                if expect_reply and device.node_desc.is_end_device in (True, None):
+                    LOGGER.debug(
+                        "Extending timeout for %s/0x%04x", device.ieee, device.nwk
+                    )
+                    await self._ezsp.setExtendedTimeout(device.ieee, True)
+                if self.use_source_routing and device.relays is not None:
+                    res = await self._ezsp.setSourceRoute(device.nwk, device.relays)
+                    if res[0] != t.EmberStatus.SUCCESS:
+                        LOGGER.warning(
+                            "Couldn't set source route for %s: %s", device.nwk, res
+                        )
+                    else:
+                        aps_frame.options = t.EmberApsOption(
+                            aps_frame.options
+                            ^ t.EmberApsOption.APS_OPTION_ENABLE_ROUTE_DISCOVERY
+                        )
+                        LOGGER.debug(
+                            "Set source route for %s to %s: %s",
+                            device.nwk,
+                            device.relays,
+                            res,
+                        )
+                res = await self._ezsp.sendUnicast(
+                    self.direct, device.nwk, aps_frame, message_tag, data
+                )
                 if res[0] != t.EmberStatus.SUCCESS:
-                    hdr, hdr_args = self._dst_pp(nwk, aps_frame)
-                    msg = hdr + "message send failure: %s"
-                    msg_args = (hdr_args + (res[0], ))
-                    raise DeliveryError(msg % msg_args)
+                    return res[0], "EZSP sendUnicast failure: %s" % (res[0],)
 
-                res = await asyncio.wait_for(req.send, timeout=APS_ACK_TIMEOUT)
-
-            if expect_reply:
-                res = await asyncio.wait_for(req.reply, timeout)
+                res = await asyncio.wait_for(req.result, APS_ACK_TIMEOUT)
         return res
 
     def permit_ncp(self, time_s=60):
@@ -380,7 +519,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if key is None:
             raise Exception("Invalid install code")
 
-        v = await self._ezsp.addTransientLinkKey(node, key)
+        link_key = t.EmberKeyData()
+        link_key.contents = key
+        v = await self._ezsp.addTransientLinkKey(node, link_key)
+
         if v[0] != t.EmberStatus.SUCCESS:
             raise Exception("Failed to set link key")
 
@@ -389,13 +531,39 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             t.EzspDecisionId.GENERATE_NEW_TC_LINK_KEY,
         )
         if v[0] != t.EmberStatus.SUCCESS:
-            raise Exception("Failed to change policy to allow generation of new trust center keys")
+            raise Exception(
+                "Failed to change policy to allow generation of new trust center keys"
+            )
 
-        return self._ezsp.permitJoining(time_s, True)
+        return await self.permit(time_s)
 
-    async def broadcast(self, profile, cluster, src_ep, dst_ep, grpid, radius,
-                        sequence, data,
-                        broadcast_address=BroadcastAddress.RX_ON_WHEN_IDLE):
+    async def broadcast(
+        self,
+        profile,
+        cluster,
+        src_ep,
+        dst_ep,
+        grpid,
+        radius,
+        sequence,
+        data,
+        broadcast_address=BroadcastAddress.RX_ON_WHEN_IDLE,
+    ):
+        """Submit and send data out as an unicast transmission.
+
+        :param profile: Zigbee Profile ID to use for outgoing message
+        :param cluster: cluster id where the message is being sent
+        :param src_ep: source endpoint id
+        :param dst_ep: destination endpoint id
+        :param: grpid: group id to address the broadcast to
+        :param radius: max radius of the broadcast
+        :param sequence: transaction sequence number of the message
+        :param data: zigbee message payload
+        :param timeout: how long to wait for transmission ACK
+        :param broadcast_address: broadcast address.
+        :returns: return a tuple of a status and an error_message. Original requestor
+                  has more context to provide a more meaningful error message
+        """
         if not self.is_controller_running:
             raise ControllerError("ApplicationController is not running")
 
@@ -404,26 +572,21 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         aps_frame.clusterId = t.uint16_t(cluster)
         aps_frame.sourceEndpoint = t.uint8_t(src_ep)
         aps_frame.destinationEndpoint = t.uint8_t(dst_ep)
-        aps_frame.options = t.EmberApsOption(
-            t.EmberApsOption.APS_OPTION_NONE
-        )
+        aps_frame.options = t.EmberApsOption(t.EmberApsOption.APS_OPTION_NONE)
         aps_frame.groupId = t.uint16_t(grpid)
         aps_frame.sequence = t.uint8_t(sequence)
+        message_tag = self.get_sequence()
 
-        with self._pending.new(sequence) as req:
+        with self._pending.new(message_tag) as req:
             async with self._in_flight_msg:
-                res = await self._ezsp.sendBroadcast(broadcast_address,
-                                                     aps_frame, radius,
-                                                     sequence, data)
+                res = await self._ezsp.sendBroadcast(
+                    broadcast_address, aps_frame, radius, message_tag, data
+                )
                 if res[0] != t.EmberStatus.SUCCESS:
-                    hdr, hdr_args = self._dst_pp(broadcast_address, aps_frame)
-                    msg = hdr + "Broadcast failure: %s"
-                    msg_args = (hdr_args + (res[0], ))
-                    raise DeliveryError(msg % msg_args)
+                    return res[0], "broadcast send failure"
 
                 # Wait for messageSentHandler message
-                res = await asyncio.wait_for(req.send,
-                                             timeout=APS_ACK_TIMEOUT)
+                res = await asyncio.wait_for(req.result, timeout=APS_ACK_TIMEOUT)
         return res
 
     async def _watchdog(self):
@@ -433,8 +596,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await asyncio.sleep(WATCHDOG_WAKE_PERIOD)
         while True:
             try:
-                await asyncio.wait_for(self.controller_event.wait(),
-                                       timeout=WATCHDOG_WAKE_PERIOD * 2)
+                await asyncio.wait_for(
+                    self.controller_event.wait(), timeout=WATCHDOG_WAKE_PERIOD * 2
+                )
                 await self._ezsp.nop()
                 failures = 0
             except (asyncio.TimeoutError, EzspError) as exc:
@@ -445,64 +609,43 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             await asyncio.sleep(WATCHDOG_WAKE_PERIOD)
 
         self._handle_reset_request(
-            "Watchdog timeout. Heartbeat timeouts: {}".format(failures))
+            "Watchdog timeout. Heartbeat timeouts: {}".format(failures)
+        )
 
+    def handle_route_record(
+        self,
+        nwk: t.EmberNodeId,
+        ieee: t.EmberEUI64,
+        lqi: t.uint8_t,
+        rssi: t.int8s,
+        relays: t.LVList(t.EmberNodeId),
+    ) -> None:
+        LOGGER.debug(
+            "Processing route record request: %s", (nwk, ieee, lqi, rssi, relays)
+        )
+        try:
+            dev = self.get_device(ieee=ieee)
+        except KeyError:
+            LOGGER.debug("Why we don't have a device for %s ieee and %s NWK", ieee, nwk)
+            self.handle_join(nwk, ieee, 0)
+            return
+        dev.relays = relays
 
-class Requests(dict):
-    def new(self, sequence, expect_reply=False):
-        """Wrap new request into a context manager."""
-        return Request(self, sequence, expect_reply)
-
-
-class Request:
-    """Context manager."""
-
-    def __init__(self, pending, sequence, expect_reply=False):
-        """Init context manager for sendUnicast/sendBroadcast."""
-        assert sequence not in pending
-        self._pending = pending
-        self._reply_fut = None
-        if expect_reply:
-            self._reply_fut = asyncio.Future()
-        self._send_fut = asyncio.Future()
-        self._sequence = sequence
-
-    @property
-    def reply(self):
-        """Reply Future."""
-        return self._reply_fut
-
-    @property
-    def sequence(self):
-        """Send Future."""
-        return self._sequence
-
-    @property
-    def send(self):
-        return self._send_fut
-
-    def __enter__(self):
-        """Return context manager."""
-        self._pending[self.sequence] = self
-        return self
-
-    def __exit__(self, exc_type, exc_value, exc_traceback):
-        """Clean up pending on exit."""
-        if not self.send.done():
-            self.send.cancel()
-        if self.reply and not self.reply.done():
-            self.reply.cancel()
-        self._pending.pop(self.sequence)
-
-        return not exc_type
+    def handle_route_error(self, status: t.EmberStatus, nwk: t.EmberNodeId) -> None:
+        LOGGER.debug("Processing route error: status=%s, nwk=%s", status, nwk)
+        try:
+            dev = self.get_device(nwk=nwk)
+        except KeyError:
+            LOGGER.debug("No %s device found", nwk)
+            return
+        dev.relays = None
 
 
 class EZSPCoordinator(CustomDevice):
     """Zigpy Device representing Coordinator."""
 
     class EZSPEndpoint(CustomEndpoint):
-        async def add_to_group(self, grp_id: int,
-                               name: str = None) -> t.EmberStatus:
+        async def add_to_group(self, grp_id: int, name: str = None) -> t.EmberStatus:
             if grp_id in self.member_of:
                 return t.EmberStatus.SUCCESS
 
@@ -530,16 +673,18 @@ class EZSPCoordinator(CustomDevice):
             return status
 
     signature = {
-        1: {
-            'profile_id': 0x0104,
-            'device_type': 0xbeef,
-            'input_clusters': [],
-            'output_clusters': [zigpy.zcl.clusters.security.IasZone.cluster_id]
+        "endpoints": {
+            1: {
+                "profile_id": 0x0104,
+                "device_type": 0xBEEF,
+                "input_clusters": [],
+                "output_clusters": [zigpy.zcl.clusters.security.IasZone.cluster_id],
+            }
         }
     }
 
     replacement = {
-        'endpoints': {
-            1: (EZSPEndpoint, {})
-        }
+        "manufacturer": "Silicon Labs",
+        "model": "EZSP",
+        "endpoints": {1: (EZSPEndpoint, {})},
     }
